(* A Sequent Calculus for BCD Subtyping with Constructors *)

Require Import Arith_base Max.
Require Import Psatz.

Require Import Eqdep_dec.


(* Basics on axiom of finite choice and product order over nat *)
Require Import AFC.
Require Import AFC_vec.
Require Import wf_prod.


(* Add-ons for List, Permutation, Vector *)
Require Import List_more.
Require Import Permutation_more.
Require Import Permutation_solve.
Require Import Vector_more.
Import ListNotations.



(** Generic System with Constructors **)

(** * Constructors with arities *)
Record Cparam := mk_Cparam {
  Ctor : Set ;
  cteqdec : forall x y:Ctor, {x=y} + {x<>y} ;
  ctvar : Ctor -> nat ;
  covar : Ctor -> nat ;
  width : Ctor -> nat ;
  valwidth : forall C, width C <= 1 }.


Section BCDgeneric.

Context {CP : Cparam}.
Notation α := (ctvar CP).
Notation β := (covar CP).
Notation ε := (width CP).

Lemma inj_pairT2_Ct : forall (P:Ctor CP -> Type) p (x y:P p),
  existT P p x = existT P p y -> x = y.
Proof.
apply inj_pair2_eq_dec.
apply cteqdec.
Qed.


(** * Intersection types with constructors *)
Inductive itype : Set :=
| constr : forall c, t itype (α c) -> t itype (β c) -> itype
| inter : itype -> itype -> itype.

Notation "s ∩ t" := (inter s t) (at level 40).
Notation "C @ a , b " := (constr C a b) (at level 40).

Fixpoint itype_ind_fix s : forall (P : itype -> Prop) (Pl : forall n, t itype n -> Prop),
  (forall C a b, Pl _ a -> Pl _ b -> P (constr C a b)) ->
  (forall s t, P s -> P t -> P (inter s t)) ->
  (Pl _ (nil _)) ->
  (forall s, P s -> forall k l, Pl k l -> Pl _ (cons _ s _ l)) ->
  P s.
Proof with try eassumption.
intros P Pl Hc Hi Hnil Hcons.
destruct s as [c t1 t2 | t1 t2].
- apply Hc.
  + induction t1...
    apply Hcons...
    eapply itype_ind_fix...
  + induction t2...
    apply Hcons...
    eapply itype_ind_fix...
- apply Hi ; eapply itype_ind_fix...
Qed.

Lemma itype_basic_ind : forall (P : itype -> Prop),
  (forall C a b, Forall P a -> Forall P b -> P (constr C a b)) ->
  (forall s t, P s -> P t -> P (inter s t)) ->
  forall s, P s.
Proof.
intros.
eapply itype_ind_fix ; try eassumption ; constructor ; assumption.
Qed.



(** * Bound on size of types *) 
Inductive isize : nat -> itype -> Prop :=
| is_c : forall C a b m n, Forall (isize m) a -> Forall (isize m) b ->
           m < n -> isize n (C@a,b)
| is_i : forall s t n, isize n s -> isize n t -> isize (S n) (s ∩ t).

Lemma isize_inc : forall s n m, isize n s -> n <= m -> isize m s.
Proof with try assumption ; try lia. 
apply (itype_basic_ind
        (fun x => forall n m, isize n x -> n <= m -> isize m x)).
- intros C a b Ha Hb n m Hn Hl.
  remember (C@a,b) as s ; revert Heqs.
  inversion Hn ; subst ; intros Heq ; inversion Heq ; subst.
  apply is_c with m0...
- intros s t Hs Ht n m Hn Hl.
  inversion Hn ; subst.
  inversion Hl ; subst...
  apply (Hs _ m0) in H2...
  apply (Ht _ m0) in H3...
  apply is_i...
Qed.

Lemma exists_isize : forall s, exists n, isize n s.
Proof with try eassumption ; try lia.
apply itype_basic_ind.
- intros C a b Ha Hb.
  apply AFCvec in Ha ;
    try (intros ; eapply isize_inc ; eassumption ; fail)...
  apply AFCvec in Hb ;
    try (intros ; eapply isize_inc ; eassumption ; fail)...
  destruct Ha as [ka Ha].
  destruct Hb as [kb Hb].
  exists (S (max ka kb)).
  apply is_c with (max ka kb)...
  + induction a ; constructor ; inversion Ha ; apply inj_pairT2_nat in H1 ; subst.
    * eapply isize_inc...
    * apply IHa...
  + induction b ; constructor ; inversion Hb ; apply inj_pairT2_nat in H1 ; subst.
    * eapply isize_inc...
    * apply IHb...
- intros s t Hs Ht.
  destruct Hs as [n1 Hs].
  destruct Ht as [n2 Ht].
  exists (S (max n1 n2)).
  apply is_i ; eapply isize_inc...
Qed.

(** * Projections for triples *)
Definition scd {A B C} : A * (B * C) -> B := fun x => fst (snd x).
Definition trd {A B C} : A * (B * C) -> C := fun x => snd (snd x).


(** * Subtyping derivation system with size parameter *)
Inductive subtype : list itype -> itype -> nat -> Prop :=
| wk : forall C a b G D s n, subtype (G ++ D) s n ->
         subtype (G ++ C@a,b :: D) s (S n)
| ir : forall G s t n, subtype G s n -> subtype G t n ->
         subtype G (s ∩ t) (S n)
| il : forall G D s t u n, subtype (G ++ s :: t :: D) u n ->
         subtype (G ++ s ∩ t :: D) u (S n)
| ct : forall C a b L n,
         ε C <= length L ->
         (forall i (Hi : i < α C),
            List.Forall (fun c => subtype [nth_order a Hi] (nth_order (fst c) Hi) n) L) ->
         (forall i (Hi : i < β C),
            subtype (List.map (fun c => (nth_order (snd c) Hi)) L) (nth_order b Hi) n) ->
         subtype (List.map (fun c => C@fst c,snd c) L) (C@a,b) (S n).

Lemma subtype_inc : forall n m G s, subtype G s n -> n <= m -> subtype G s m.
Proof with try lia ; try assumption.
induction n using (well_founded_induction lt_wf) ; intros m G s Hm Hl.
inversion Hm ; subst ; try (inversion Hl ; constructor ; apply (H n0))...
inversion Hl ; apply ct...
- intros i Hi.
  assert (HF := H1 i Hi).
  clear H0 H1 H2 H4 Hl Hm.
  induction L ; constructor ; inversion HF.
  + apply (H n0)...
  + apply IHL...
- intros i Hi.
  assert (Hs := H2 i Hi).
  apply (H n0)...
Qed.


(** * Admissible rules *)
Lemma ex : forall D G s n, subtype G s n -> Permutation G D -> subtype D s n.
Proof with try eassumption.
intros D G s n H ; revert D ; induction H ; intros G' HP.
- destruct (Permutation_vs_elt_inv _ _ _ _ (Permutation_sym HP)) as (l' & l'' & Heq) ; subst.
  apply Permutation_app_inv in HP.
  apply IHsubtype in HP.
  apply wk...
- assert (HP2 := HP).
  apply IHsubtype1 in HP.
  apply IHsubtype2 in HP2.
  apply ir...
- destruct (Permutation_vs_elt_inv _ _ _ _ (Permutation_sym HP)) as (l' & l'' & Heq) ; subst.
  apply Permutation_app_inv in HP.
  apply (Permutation_elt t) in HP.
  apply (Permutation_elt s) in HP.
  apply IHsubtype in HP.
  apply il...
- destruct (Permutation_map_inv _ _ _ (Permutation_sym HP)) as (L' & Heq & HP') ; subst.
  apply ct...
  + apply Permutation_length in HP' ; lia.
  + intros i Hi.
    specialize H0 with i Hi.
    eapply Permutation_Forall...
  + intros i Hi.
    apply H2...
    apply Permutation_map...
Qed.

Lemma wk_gen : forall t G D s n, subtype (G ++ D) s n -> exists m, subtype (G ++ t :: D) s m.
Proof with try eassumption.
induction t ; intros.
- eexists ; apply wk...
- destruct (IHt2 _ _ _ _ H) as [m2 H2].
  destruct (IHt1 _ _ _ _ H2) as [m1 H1].
  eexists ; apply il...
Qed.

Lemma ax_size : forall n s, isize n s -> exists m, subtype [s] s m.
Proof with try eassumption ; try lia.
induction n using (well_founded_induction lt_wf) ; intros s Hsize.
inversion Hsize ; subst.
- pose proof (H _ H2) as Hsub.
  apply (Forall_impl _ _ Hsub) in H0.
  apply AFCvec in H0 ; [ | intros ; eapply subtype_inc ; eassumption ].
  destruct H0 as [ka Ha].
  apply (Forall_impl _ _ Hsub) in H1.
  apply AFCvec in H1 ; [ | intros ; eapply subtype_inc ; eassumption ].
  destruct H1 as [kb Hb].
  exists (S (max ka kb)).
  change [C@a,b] with (List.map (fun c => C @ fst c, snd c) [(a,b)]).
  apply ct.
  + apply valwidth.
  + intros i Hi.
    repeat constructor ; simpl.
    eapply subtype_inc.
    * apply (Forall_nth_order _ _ _ Ha).
    * lia.
  + intros i Hi.
    repeat constructor ; simpl.
    eapply subtype_inc.
    * apply (Forall_nth_order _ _ _ Hb).
    * lia.
- assert (n0 < S n0) as Hl...
  destruct (H n0 Hl s0 H0) as [k1 Hs1].
  destruct (H n0 Hl t  H1) as [k2 Hs2].
  rewrite <- (app_nil_r [s0]) in Hs1.
  destruct (wk_gen t _ _ _ _ Hs1) as [k1' Hs1'].
  rewrite <- (app_nil_l [t]) in Hs2.
  destruct (wk_gen s0 _ _ _ _ Hs2) as [k2' Hs2'].
  exists (S (S (S (max k1' k2')))).
  rewrite <- (app_nil_l [s0 ∩ t]).
  apply il.
  apply ir ; eapply subtype_inc...
Qed.

Lemma ax : forall s, exists n, subtype [s] s n.
Proof.
intro s.
destruct (exists_isize s).
eapply ax_size ; eassumption.
Qed.

Lemma il_inv : forall n G D s t u, subtype (G ++ s ∩ t :: D) u n ->
  exists m, subtype (G ++ s :: t :: D) u m.
Proof with try eassumption ; try lia.
induction n using (well_founded_induction lt_wf) ; intros G D s t u Hi.
inversion Hi ; subst.
- dichot_elt_app_exec H0 ; subst.
  + rewrite app_assoc in H1.
    apply H in H1...
    destruct H1 as [s' H1].
    rewrite <- app_assoc in H1.
    rewrite <- app_assoc.
    exists (S s') ; apply wk...
  + destruct l0 ; inversion H3 ; subst.
    rewrite <- ? app_assoc in H1.
    rewrite <- ? app_comm_cons in H1.
    apply H in H1...
    destruct H1 as [s' H1].
    rewrite ? app_comm_cons in H1.
    rewrite app_assoc in H1.
    rewrite ? app_comm_cons.
    rewrite app_assoc.
    exists (S s') ; apply wk...
- eapply H in H0...
  eapply H in H1...
  destruct H0 as [nl Hl].
  destruct H1 as [nr Hr].
  exists (S (max nl nr)) ; apply ir ; eapply subtype_inc...
- dichot_elt_app_exec H0 ; subst.
  + rewrite ? app_comm_cons in H1.
    rewrite app_assoc in H1.
    apply H in H1...
    destruct H1 as [s' H1].
    rewrite <- ? app_assoc in H1.
    rewrite <- ? app_comm_cons in H1.
    rewrite <- ? app_assoc.
    rewrite <- ? app_comm_cons.
    exists (S s') ; apply il...
  + destruct l0 ; inversion H3 ; subst.
    * rewrite app_nil_r in H1.
      eexists...
    * rewrite <- ? app_assoc in H1.
      rewrite <- ? app_comm_cons in H1.
      apply H in H1...
      destruct H1 as [s' H1].
      rewrite ? app_comm_cons in H1.
      rewrite app_assoc in H1.
      rewrite ? app_comm_cons.
      rewrite app_assoc.
      exists (S s') ; apply il...
- exfalso.
  clear - H0 ; revert G H0 ; induction L ; intros G H ; destruct G ; inversion H.
  apply IHL in H2...
Qed.

Lemma ctr_size : forall v s, isize (fst v) s ->
  forall G u, subtype (s :: s :: G) u (snd v) -> exists m, subtype (s :: G) u m.
Proof with try lt_nat_nat_solve ; try eassumption ; try reflexivity ; try lia.
induction v as [(f,n) IH] using (well_founded_induction wf_nat_nat) ; simpl ; intros s Hs G u pi.
inversion pi ; subst.
- destruct G0 ; [ | destruct G0 ] ; inversion H ; subst.
  + eexists...
  + eexists...
  + rewrite <- ? app_comm_cons in H0.
    apply (IH (f,n0)) in H0...
    destruct H0 as [m H0].
    rewrite app_comm_cons.
    eexists ; apply wk...
- apply (IH (f,n0)) in H...
  destruct H as [m1 H].
  apply (IH (f,n0)) in H0...
  destruct H0 as [m2 H0].
  exists (S (max m1 m2)) ; apply ir ; eapply subtype_inc...
- destruct G0 ; [ | destruct G0 ] ; inversion H ; subst.
  + assert (Permutation ([] ++ s0 :: t :: s0 ∩ t :: G)
                        ((s0 :: []) ++ s0 ∩ t :: t :: G))
      as HP by perm_solve.
    eapply ex in HP...
    destruct (il_inv _ _ _ _ _ _ HP) as [m Hsub] ; simpl in Hsub.
    inversion Hs ; subst.
    apply (IH (n,m)) in Hsub...
    destruct Hsub as [m1 Hsub].
    assert (Permutation (s0 :: t :: t :: G) (t :: t :: s0 :: G))
      as HP' by perm_solve.
    eapply ex in HP'...
    apply (IH (n,m1)) in HP'...
    destruct HP' as [m2 Hsub2].
    eapply ex in Hsub2 ; [ | apply perm_swap ].
    rewrite <- (app_nil_l (_ :: G)).
    eexists ; apply il...
  + simpl in H0.
    rewrite <- (app_nil_l (_ :: _)) in H0.
    destruct (il_inv _ _ _ _ _ _ H0) as [m1 Hsub1] ; simpl in Hsub1.
    inversion Hs ; subst.
    assert (Permutation ([] ++ s0 :: t :: s0 :: t :: G)
                        (s0 :: s0 :: t :: t :: G))
      as HP by perm_solve.
    eapply (ex _ _ _ _ Hsub1) in HP.
    apply (IH (n,m1)) in HP...
    destruct HP as [m2 Hsub2].
    assert (Permutation (s0 :: t :: t :: G) (t :: t :: s0 :: G))
      as HP by perm_solve.
    eapply ex in HP...
    apply (IH (n,m2)) in HP...
    destruct HP as [m3 Hsub3].
    eapply ex in Hsub3 ; [ | apply perm_swap ].
    rewrite <- (app_nil_l (_ :: G)).
    eexists ; apply il...
  + rewrite <- ? app_comm_cons in H0.
    apply (IH (f,n0)) in H0...
    destruct H0 as [m H0].
    rewrite app_comm_cons.
    eexists ; apply il...
- symmetry in H ; decomp_map H ; subst.
  inversion H6 ; apply inj_pairT2_Ct in H3 ; apply inj_pairT2_Ct in H4.
  assert (forall i (Hi : i < β C), exists n1,
          subtype (List.map (fun c => nth_order (snd c) Hi) (x0 :: l1))
                  (nth_order b Hi) n1)
    as Hb.
  { intros i Hi.
    assert (HI := H2 i Hi) ; simpl in HI.
    change n0 with (snd (f,n0)) in HI.
    rewrite H4 in HI.
    apply IH in HI...
    simpl ; rewrite <- H4.
    simpl in Hs ; inversion Hs ;
      apply inj_pairT2_Ct in H5 ; apply inj_pairT2_Ct in H7 ; subst.
    apply Forall_nth_order.
    remember (snd x) as l'.
    revert H10 H11 ; clear ; induction l' ; intros Hs Hl ;
      inversion Hs ; subst ; constructor ; 
      [ eapply isize_inc | apply IHl' ]...
    apply inj_pairT2_nat in H1 ; subst... }
  apply AFCvec_incdep in Hb ; [ | | intros ; eapply subtype_inc ]...
  + destruct Hb as [k Hb].
    rewrite H3 ; rewrite H4.
    change (C @ fst x0, snd x0
             :: List.map (fun c => C @ fst c, snd c) l1)
      with (List.map (fun c => C @ fst c, snd c) (x0::l1)).
    eexists (S (max n0 k)) ; apply ct.
    * simpl ; pose proof (valwidth CP C) ; lia.
    * intros i Hi.
      assert (H1' := H1 i Hi).
      inversion H1'.
      inversion H8.
      constructor.
      -- eapply subtype_inc...
      -- revert H12 ; clear ; induction l1 ; intro H ; constructor ; inversion H ;
           [ eapply subtype_inc | apply IHl1]...
    * intros i Hi.
      eapply subtype_inc ; [ apply Hb | lia].
  + remember (x0 :: l1) as l ; clear.
    intros i n H1 H2.
    enough (forall l0, subtype (List.map (fun c => nth_order (snd c) H2) l0
                             ++ List.map (fun c => nth_order (snd c) H1) l) (nth_order b H1) i ->
                       subtype (List.map (fun c => nth_order (snd c) H2) (l0 ++ l)) (nth_order b H2) i)
      as H0 by (apply (H0 [])).
    induction l ; intros l0 HI ; simpl in HI ; simpl.
    * rewrite <- (nth_order_ext _ H1).
      rewrite app_nil_r ; rewrite app_nil_r in HI...
    * replace (l0 ++ a :: l) with ((l0 ++ [a]) ++ l) by (rewrite <- app_assoc ; reflexivity).
      apply IHl.
      rewrite map_app ; simpl.
      rewrite <- (nth_order_ext _ H1).
      rewrite <- app_assoc...
Qed.

Lemma ctr : forall G s u n, subtype (s :: s :: G) u n -> exists m, subtype (s :: G) u m.
Proof.
intros.
destruct (exists_isize s) as [f Hs].
change f with (fst (f,n)) in Hs.
change n with (snd (f,n)) in H.
apply ctr_size in H ; eassumption.
Qed.


Theorem cut_size : forall v s, isize (fst v) s ->
forall D G t, subtype D s (scd v) -> subtype (s :: G) t (trd v) ->
  exists m, subtype (D ++ G) t m.
Proof with try lt_nat_nat_nat_solve ;
           try eassumption ; try reflexivity ; try lia ; try perm_solve.
induction v as [(f,(l,r)) IH] using (well_founded_induction wf_nat_nat_nat).
intros s Hs D G t pi1 pi2.
inversion pi1 ; subst.
- unfold scd in H2 ; simpl in H2 ; subst.
  change n with (scd (f,(n,r))) in H.
  eapply IH in H...
  destruct H as [m H].
  rewrite <- ? app_assoc in H.
  rewrite <- ? app_assoc ; rewrite <- app_comm_cons.
  eexists ; apply wk...
- inversion pi2 ; subst.
  + unfold scd in H3 ; simpl in H3 ; subst.
    unfold trd in H5 ; simpl in H5 ; subst.
    destruct G0 ; inversion H1 ; subst.
    rewrite <- app_comm_cons in H2.
    change n0 with (trd (f, (S n, n0))) in H2.
    change (scd (f, (S n, S n0))) with (scd (f, (S n, n0))) in pi1.
    eapply IH in H2...
    destruct H2 as [m HI].
    rewrite app_assoc in HI.
    rewrite app_assoc.
    eexists ; apply wk...
  + unfold scd in H3 ; simpl in H3 ; subst.
    unfold trd in H6 ; simpl in H6 ; subst.
    change n0 with (trd (f,(S n,n0))) in H1.
    change n0 with (trd (f,(S n,n0))) in H2.
    change (scd (f,(S n, S n0))) with (scd (f,(S n,n0))) in pi1.
    eapply IH in H1...
    eapply IH in H2...
    destruct H1 as [m1 HI1].
    destruct H2 as [m2 HI2].
    eexists (S (max m1 m2)) ; apply ir ; eapply subtype_inc...
  + unfold scd in H3 ; simpl in H3 ; subst.
    unfold trd in H5 ; simpl in H5 ; subst.
    inversion Hs ; subst.
    destruct G0 ; inversion H1 ; subst.
    * simpl in H5 ; subst.
      change n0 with (trd (n1, (n, n0))) in H2.
      change n  with (scd (n1, (n, n0))) in H.
      eapply IH in H2...
      destruct H2 as [m H2].
      change n with (scd (n1, (n, m))) in H0.
      change m with (trd (n1, (n, m))) in H2.
      eapply ex in H2 ; [ | symmetry ; apply Permutation_cons_app ; reflexivity ].
      eapply IH in H2...
      destruct H2 as [m' HI].
      clear - HI ; revert m' G HI ; induction D ; intros m G HI.
      -- eexists...
      -- rewrite <- ? app_comm_cons in HI.
         assert (Permutation (a :: D ++ a :: D ++ G) (a :: a :: D ++ D ++ G))
           as HP by perm_solve.
         apply (ex _ _ _ _ HI) in HP...
         destruct (ctr _ _ _ _ HP) as [m1 HP1].
         eapply ex in HP1.
         ++ apply IHD in HP1.
            destruct HP1.
            rewrite <- ? app_comm_cons.
            eexists ; eapply ex ; [ | symmetry ; apply Permutation_cons_app ]...
         ++ perm_solve.
    * change (scd (f, (S n, S n0))) with (scd (f, (S n, n0))) in pi1.
      change n0 with (trd (f,(S n,n0))) in H2.
      rewrite <- app_comm_cons in H2.
      eapply IH in H2...
      destruct H2 as [m' H2].
      rewrite app_assoc in H2.
      rewrite app_assoc.
      eexists ; apply il...
  + exfalso.
    symmetry in H1 ; decomp_map H1.
    inversion H8.
- unfold scd in H2 ; simpl in H2 ; subst.
  change n with (scd (f, (n, r))) in H.
  eapply IH in H...
  destruct H as [m H].
  rewrite <- app_assoc in H ; rewrite <- ? app_comm_cons in H.
  rewrite <- app_assoc ; rewrite <- app_comm_cons.
  eexists ; apply il...
- inversion pi2 ; subst.
  + unfold scd in H4 ; simpl in H4 ; subst.
    unfold trd in H6 ; simpl in H6 ; subst.
    destruct G0 ; inversion H2 ; subst.
    * apply inj_pairT2_Ct in H6 ; apply inj_pairT2_Ct in H7 ; subst.
      revert H3 ; clear ; induction L ; intro HI.
      -- eexists ; apply HI.
      -- apply IHL in HI...
         destruct HI as [m HI].
         simpl ; rewrite <- (app_nil_l (constr _ _ _ :: _)).
         eexists ; apply wk...
    * rewrite <- app_comm_cons in H3.
      change n0 with (trd (f,(S n,n0))) in H3.
      change (scd (f,(S n, S n0))) with (scd (f,(S n,n0))) in pi1.
      eapply IH in H3...
      destruct H3 as [m HI].
      rewrite app_assoc in HI.
      rewrite app_assoc.
      eexists ; apply wk...
  + unfold scd in H4 ; simpl in H4 ; subst.
    unfold trd in H7 ; simpl in H7 ; subst.
    change n0 with (trd (f, (S n, n0))) in H2.
    change n0 with (trd (f, (S n, n0))) in H3.
    change (scd (f, (S n, S n0))) with (scd (f, (S n,n0))) in pi1.
    eapply IH in H2...
    eapply IH in H3...
    destruct H2 as [m1 HI1].
    destruct H3 as [m2 HI2].
    eexists (S (max m1 m2)) ; apply ir ; eapply subtype_inc...
  + unfold scd in H4 ; simpl in H4 ; subst.
    unfold trd in H6 ; simpl in H6 ; subst.
    destruct G0 ; inversion H2 ; subst.
    change n0 with (trd (f, (S n, n0))) in H3.
    change (scd (f, (S n, S n0))) with (scd (f, (S n, n0))) in pi1.
    rewrite <- app_comm_cons in H3.
    eapply IH in H3...
    destruct H3 as [m HI].
    rewrite app_assoc in HI.
    rewrite app_assoc.
    eexists ; apply il...
  + unfold scd in H4 ; simpl in H4 ; subst.
    unfold trd in H8 ; simpl in H8 ; subst.
    symmetry in H2 ; decomp_map H2 ; inversion H7 ; subst ;
      apply inj_pairT2_Ct in H10 ; apply inj_pairT2_Ct in H11 ; subst.
    assert (forall i (Hi : i < α C0), exists k, 
            List.Forall (fun c => subtype [nth_order a0 Hi]
                                          (nth_order (fst c) Hi) k)
              (L ++ l0)) as Ha.
    { intros i Hi.
      apply AFClist ; [ intros ; eapply subtype_inc ; eassumption | ].
      assert (Hl2 := H5 i Hi).
      inversion Hl2 ; subst.
      apply Forall_app.
      - assert (HL := H0 i Hi).
        revert HL ; clear - IH Hs Hi H8 ; induction L ; intro HL ;
          constructor ; inversion HL ; subst.
        + inversion Hs ; apply inj_pairT2_Ct in H0 ; apply inj_pairT2_Ct in H3 ; subst.
          simpl in H7.
          change n0 with (scd (m, (n0, n))) in H8.
          change n  with (trd (m, (n0, n))) in H2.
          eapply IH in H8 ; simpl in H8...
          apply Forall_nth_order...
        + apply IHL...
      - apply exists_Forall.
        eexists... }
    apply AFCvec_incdep in Ha.
    * destruct Ha as [ka Ha].
      assert (forall i (Hi : i < β C0), exists k,
                subtype (List.map (fun c => nth_order (snd c) Hi) (L ++ l0))
                        (nth_order b0 Hi) k) as Hb.
      { intros i Hi.
        assert (HL := H1 i Hi).
        assert (HR := H6 i Hi).
        inversion Hs ; apply inj_pairT2_Ct in H4 ; apply inj_pairT2_Ct in H8 ; subst.
        simpl in H12.
        change n  with (scd (m, (n, n0))) in HL.
        change n0 with (trd (m, (n, n0))) in HR.
        rewrite ? map_app.
        eapply IH in HL...
        apply Forall_nth_order... }
      apply AFCvec_incdep in Hb ; [ | | intros ; eapply subtype_inc ; eassumption ].
      -- destruct Hb as [kb Hb].
         rewrite <- map_app.
         exists (S (max (max n n0) (max ka kb))) ; apply ct...
         ++ rewrite app_length...
         ++ intros i0 Hi0.
            assert (Ha' := Ha i0 Hi0).
            eapply (List_more.inc_Forall
                     (fun i c => subtype [nth_order a0 Hi0] (nth_order (fst c) Hi0) i))...
            intros ; eapply subtype_inc...
         ++ intros i0 Hi0.
            assert (Hb' := Hb i0 Hi0).
            eapply subtype_inc...
      -- remember (L ++ l0) as l ; clear ; intros i n H1 H2.
         enough (forall l0, subtype (List.map (fun c => nth_order (snd c) H2) l0
                                  ++ List.map (fun c => nth_order (snd c) H1) l) (nth_order b0 H1) i ->
                       subtype (List.map (fun c => nth_order (snd c) H2) (l0 ++ l)) (nth_order b0 H2) i)
           as H0 by (apply (H0 [])).
         induction l ; intros l0 HI ; simpl in HI ; simpl.
         ++ rewrite <- (nth_order_ext _ H1).
            rewrite app_nil_r ; rewrite app_nil_r in HI...
         ++ replace (l0 ++ a :: l) with ((l0 ++ [a]) ++ l) by (rewrite <- app_assoc ; reflexivity).
            apply IHl.
            rewrite map_app ; simpl.
            rewrite <- (nth_order_ext _ H1).
            rewrite <- app_assoc...
    * intros.
      eapply (List_more.inc_Forall
               (fun i c => subtype [nth_order a0 H4] (nth_order (fst c) H4) i))...
      -- intros ; eapply subtype_inc...
      -- remember (L ++ l0) as l ; clear - H8.
         revert H8 ; induction l ; intros HI ; simpl in HI ; simpl ; try now constructor.
         inversion HI ; subst ; constructor.
         ++ rewrite 2 (nth_order_ext _ H4 H2)...
         ++ apply IHl...
    * remember (L ++ l0) as l ; clear ; induction l ; intros ; try now constructor.
      inversion H0 ; constructor ; subst.
      -- eapply subtype_inc...
      -- eapply IHl...
Qed.

Theorem cut : forall D s G t n m,
  subtype D s n -> subtype (s :: G) t m -> exists k, subtype (D ++ G) t k.
Proof.
intros.
destruct (exists_isize s) as [f Hs].
change f with (fst (f,(n,m))) in Hs.
change n with (scd (f,(n,m))) in H.
change m with (trd (f,(n,m))) in H0.
eapply cut_size ; eassumption.
Qed.

End BCDgeneric.





(** BCD presentation *)
Section BCDstd.

Variable BCtor : Set.
Variable beqdec : forall x y : BCtor, {x=y} + {x<>y}.
Variable bvar : BCtor -> nat.

Definition BCDCtor := (option (option BCtor)).
  (* None is Omega, Some None is Arrow *)
Fixpoint bctvar (C:BCDCtor) :=
  (* contravariance is 0 except for Arrow *)
  match C with
  | Some None => 1
  | _ => 0
  end.
Fixpoint bcovar C :=
  (* cotravariance is 0 for Omega and 1 for Arrow *)
  match C with
  | None => 0
  | Some None => 1
  | Some (Some c) => bvar c
  end.
Fixpoint bwidth (C:BCDCtor) :=
  (* width is 0 for Omega and Arrow and 1 for constructors *)
  match C with
  | None => 0
  | Some None => 0
  | _ => 1
  end.

Lemma bcdeqdec : forall x y : BCDCtor, {x=y} + {x<>y}.
Proof with try reflexivity ;
           try (left ; reflexivity) ;
           try (right ; intros Heq ; inversion Heq ; fail).
intros.
destruct x as [x | ] ; destruct y as [y | ]...
destruct x as [x | ] ; destruct y  as [y | ]...
destruct (beqdec x y) as [ Heq | Hneq ] ; [ left | right ] ; subst...
intros Heq ; apply Hneq ; inversion Heq...
Qed.

Lemma bvalwidth : forall C, bwidth C <= 1.
Proof.
intros C ; destruct C as [ C | ] ; [ destruct C | ] ; simpl ; lia.
Qed.

Definition BCDC := mk_Cparam BCDCtor bcdeqdec bctvar bcovar bwidth bvalwidth.
Notation α := (ctvar BCDC).
Notation β := (covar BCDC).
Notation ε := (width BCDC).
Notation "s ∩ t" := (inter s t) (at level 40).
Notation "C @ a , b " := (@constr BCDC C a b) (at level 40).
Notation Ω := (None@nil _,nil _).
Notation "s → t" := ((Some None)@cons _ s _ (nil _), cons _ t _ (nil _)) (at level 42).
Notation κ := (fun c a => ((Some (Some c)) @ nil _ , a)).

Reserved Notation "a ≤ b " (at level 70).
Inductive sub_bcd : @itype BCDC -> @itype BCDC -> Prop :=
| id_bcd          : forall a, a ≤ a
| trans_bcd       : forall a b c, a ≤ b -> b ≤ c -> a ≤ c
| omega_right_bcd : forall a, a ≤ Ω
| inter_left1_bcd : forall a b, a ∩ b ≤ a
| inter_left2_bcd : forall a b, a ∩ b ≤ b
| inter_right_bcd : forall a, a ≤ a ∩ a
| inter_bcd       : forall a b c d, a ≤ c -> b ≤ d -> a ∩ b ≤ c ∩ d
| arrow_bcd       : forall a b c d, c ≤ a -> b ≤ d -> a → b ≤ c → d
| arrow_inter_bcd : forall a b c, (a → b) ∩ (a → c) ≤ a → b ∩ c
| arrow_omega_bcd : Ω ≤ Ω → Ω
| ctor_mon_bcd    : forall C p l b, nth l p ≤ b -> κ C l ≤ κ C (replace l p b)
| ctor_inter_bcd  : forall C l1 l2, κ C l1 ∩ κ C l2 ≤ κ C (map2 inter l1 l2)
where "a ≤ b" := (sub_bcd a b).

Lemma ctor_mon_bcd_gen : forall C l1 l2, Forall2 (fun s t => s ≤ t) l1 l2 -> κ C l1 ≤ κ C l2.
Proof with try lia ; try eassumption ; try reflexivity.
intros C l1 l2 HF.
assert (forall p (Hp : p <= β (Some (Some C))) l, (forall k1 (Hk1 : k1 < p),
  nth_order l (lt_le_trans _ _ _ Hk1 Hp) = nth_order l2 (lt_le_trans _ _ _ Hk1 Hp)) ->
  (forall k2 (Hk2 : k2 < β (Some (Some C))), p <= k2 ->
  nth_order l Hk2 = nth_order l1 Hk2) -> κ C l1 ≤ κ C l) as HI.
{ induction p ; intros.
  - enough (l = l1) by (subst ; apply id_bcd).
    clear H ; revert l H0 ; clear ; revert l1.
    induction (β (Some (Some C))) ; intros.
    + apply case0.
      apply (case0_nil l)...
    + rewrite (eta l1) ; rewrite (eta l) ; f_equal.
      * assert (0 < S n) as HS...
        rewrite 2 (hd_nth_order HS).
        apply H0...
      * apply IHn ; intros.
        assert (S k2 < S n) as Hk2S...
        rewrite 2 (nth_order_tl _ Hk2 Hk2S).
        apply H0...
  - assert (p < β (Some (Some C))) as Hp'...
    apply (trans_bcd _ (Some (Some C) @ nil itype,
                         (replace l (Fin.of_nat_lt Hp') (nth_order l1 Hp')))).
    + assert (p <= β (Some (Some C))) as Hple...
      apply (IHp Hple) ; intros.
      * rewrite nth_order_replace_neq...
        assert (k1 < S p) as Hk1S...
        rewrite 2 (nth_order_ext _ (Nat.lt_le_trans k1 p (β (Some (Some C))) Hk1 Hple)
                                   (Nat.lt_le_trans _ _ _ Hk1S Hp)).
        apply H.
      * inversion H1 ; subst.
        -- rewrite nth_order_replace_eq.
           apply nth_order_ext.
        -- transitivity (nth_order l Hk2).
           ++ rewrite (nth_order_replace_neq)...
           ++ apply H0...
    + replace l with (replace (replace l (Fin.of_nat_lt Hp') (nth_order l1 Hp'))
                              (Fin.of_nat_lt Hp') (nth_order l Hp')) at 2.
      * apply ctor_mon_bcd.
        change (nth (replace l (Fin.of_nat_lt Hp') (nth_order l1 Hp')) (Fin.of_nat_lt Hp'))
          with (nth_order (replace l (Fin.of_nat_lt Hp') (nth_order l1 Hp')) Hp').
        rewrite nth_order_replace_eq.
        replace (nth_order l Hp') with (nth_order l2 Hp').
        -- apply Forall2_nth_order...
        -- symmetry.
           assert (p < S p) as HpS...
           rewrite 2 (nth_order_ext _ Hp' (lt_le_trans _ _ _ HpS Hp)).
           apply H.
      * rewrite replace_replace_eq.
        unfold nth_order ; rewrite replace_id... }
apply (HI (β (Some (Some C))) (le_refl _)) ; intros...
exfalso...
Qed.

(** * Correctness with respect to BCD *)
Theorem correctness : forall a b, a ≤ b -> exists n, subtype [a] b n.
Proof with try eassumption ; try reflexivity ; try lia.
intros a b Hbcd.
induction Hbcd ;
  try destruct (ax a) as [na Haxa] ;
  try destruct (ax b) as [nb Haxb] ;
  try destruct (ax c) as [nc Haxc] ;
  try (destruct IHHbcd1 as [nl Hl]) ;
  try (destruct IHHbcd2 as [nr Hr])...
- apply ax...
- rewrite <- (app_nil_r [a]).
  clear Haxa Haxb ; eapply cut...
- rewrite <- (app_nil_l [a]).
  eapply (wk_gen _ _ _ _ 1)...
  change [] with (List.map (fun x => (None@fst x,snd x)) []).
  apply ct ; try constructor...
  intros i Hi ; inversion Hi.
- rewrite <- (app_nil_l [_]) in Haxa.
  destruct (wk_gen b _ _ _ _ Haxa) as [m Hax].
  rewrite <- (app_nil_l [_]).
  eexists ; apply il.
  apply (ex _ [b;a])...
  apply Permutation_rev.
- rewrite <- (app_nil_l [_]) in Haxb.
  destruct (wk_gen a _ _ _ _ Haxb) as [m Hax].
  rewrite <- (app_nil_l [_]).
  eexists ; apply il...
- eexists ; apply ir...
- rewrite <- (app_nil_r [_]) in Hl.
  destruct (wk_gen b _ _ _ _ Hl) as [m1 H1].
  rewrite <- (app_nil_l [_]) in Hr.
  destruct (wk_gen a _ _ _ _ Hr) as [m2 H2].
  exists (S (S (S (max m1 m2)))).
  rewrite <- (app_nil_l [_]).
  apply il ; apply ir ; eapply subtype_inc...
- change [a → b] with
    (List.map (fun x => Some None @ fst x , snd x) [(cons _ a _ (nil _), cons _ b _ (nil _))]).
  exists (S (max nl nr)) ; apply ct ; simpl...
  + intros i Hi ; constructor ; [ | constructor].
    assert (i = 0) as H0 by lia ; subst.
    eapply subtype_inc...
  + intros i Hi.
    assert (i = 0) as H0 by lia ; subst.
    eapply subtype_inc...
- rewrite <- (app_nil_r [_]) in Haxb.
  destruct (wk_gen c _ _ _ _ Haxb) as [mb Hb].
  rewrite <- (app_nil_l [_]) in Haxc.
  destruct (wk_gen b _ _ _ _ Haxc) as [mc Hc].
  rewrite <- (app_nil_l [_]).
  eexists ; apply il.
  change [a → b ; a → c] with
    (List.map (fun x => Some None @ fst x , snd x) [(cons _ a _ (nil _),
                                                     cons _ b _ (nil _));
                                                    (cons _ a _ (nil _),
                                                     cons _ c _ (nil _))]).
  apply ct ; simpl...
  + intros i Hi ; constructor ; [ | constructor ; [ | constructor] ].
    * assert (i = 0) as H0 by lia ; subst.
      apply (subtype_inc (S (S (max na (max mb mc)))))...
      eapply subtype_inc...
    * assert (i = 0) as H0 by lia ; subst.
      apply (subtype_inc (S (S (max na (max mb mc)))))...
      eapply subtype_inc...
  + intros i Hi.
    assert (i = 0) as H0 by lia ; subst.
    apply ir ; eapply subtype_inc...
- rewrite <- (app_nil_l [_]).
  exists 3 ; apply wk...
  change [] with (List.map (fun x => Some None @ fst x , snd x) []).
  apply ct ; simpl ; intros ; try constructor...
  assert (i = 0) as H0 by lia ; subst.
  change [] with (List.map (fun x => None @ fst x , snd x) []).
  apply ct ; intros ; try constructor...
  inversion Hi0.
- assert (Forall (fun x => exists k, subtype [x] x k) l) as Hl.
  { clear ; induction l ; constructor.
    - apply ax...
    - apply IHl... }
  apply AFCvec in Hl ; [ | intros ; eapply subtype_inc ; eassumption].
  destruct Hl as [n Hl].
  destruct IHHbcd as [n' Hab].
  change [Some (Some C) @ nil _ , l] with
    (List.map (fun x => Some (Some C) @ fst x , snd x) [(nil _,l)]).
  exists (S (max n' n)) ; apply ct ; simpl...
  + intros i Hi ; inversion Hi.
  + intros i Hi.
    assert (exists k, bvar C = S k) as [k Hk].
    { destruct (bvar C) ; [exfalso | exists n0]... }
    revert p n n' l Hbcd Hab Hl Hi ; simpl ; rewrite Hk ;
      clear ; revert k ; induction i ; intros k p ; destruct p ; intros.
    * rewrite (eta l) ; rewrite (eta l) in Hab.
      rewrite <- 2 hd_nth_order ; simpl ; simpl in Hab.
      eapply subtype_inc...
    * rewrite (eta l) ; rewrite (eta l) in Hl.
      rewrite <- 2 hd_nth_order ; simpl.
      inversion Hl ; subst.
      eapply subtype_inc...
   * rewrite (eta l) ; simpl.
     assert (i < n) as Hi'...
     rewrite <- 2 (nth_order_tl _ Hi') ; simpl.
     rewrite (nth_order_tl _ _ Hi).
     eapply Forall_nth_order in Hl.
     eapply subtype_inc...
   * rewrite (eta l) ; simpl.
     assert (i < n) as Hi'...
     rewrite <- 2 (nth_order_tl _ Hi') ; simpl.
     assert (exists k, n = S k) as [k' Hk].
     { clear - Hi ; destruct n ; [exfalso | exists n]... }
     revert p l Hbcd Hab Hl Hi Hi' ; simpl ; rewrite Hk ; intros.
     eapply (IHi k').
     -- rewrite (eta l) in Hbcd...
     -- rewrite (eta l) in Hab...
     -- rewrite (eta l) in Hl.
        inversion Hl ; subst.
        apply inj_pairT2_nat in H1 ; subst...
- assert (forall k (l : t _ k) k' (l' : t _ k'),
             Forall (fun x => exists n, Forall (fun y => @subtype BCDC [y;x] x n) l') l)
    as HAXl.
  { clear ; induction l ; constructor.
    - apply AFCvec.
      + intros ; eapply subtype_inc...
      + induction l' ; constructor.
        * destruct (ax h) as [k Hax].
          rewrite <- (app_nil_l [h0; h]).
          eapply wk_gen...
        * apply IHl'...
    - apply IHl... }
  assert (HAX1 := HAXl _ l1 _ l2).
  apply AFCvec in HAX1...
  destruct HAX1 as [n1 HAX1].
  assert (HAX2 := HAXl _ l2 _ l1).
  apply AFCvec in HAX2...
  destruct HAX2 as [n2 HAX2].
  rewrite <- (app_nil_l [_]).
  exists (S (S (S (S (max n1 n2))))) ; apply il.
  change [Some (Some C) @ nil _ , l1 ; Some (Some C) @ nil _ , l2]
    with (List.map (fun x => Some (Some C) @ fst x , snd x) [(nil _,l1);(nil _,l2)]).
  apply ct ; simpl...
  + intros i Hi ; inversion Hi.
  + intros i Hi.
    assert (exists k, bvar C = S k) as [k Hk].
    { destruct (bvar C) ; [exfalso | exists n]... }
    revert HAXl n1 n2 l1 l2 HAX1 HAX2 Hi ; simpl ; rewrite Hk ; clear Hk ; intros HAXl n1 n2.
    revert k ; induction i ; intros k l1 l2 HAX1 HAX2 Hi...
    * rewrite (eta l1) ; rewrite (eta l2).
      rewrite <- ? hd_nth_order ; simpl.
      apply ir.
      -- rewrite (eta l1) in HAX1 ; inversion HAX1.
         apply inj_pairT2_nat in H1 ; subst.
         rewrite (eta l2) in H2 ; inversion H2.
         apply inj_pairT2_nat in H1 ; subst.
         eapply ex.
         ++ eapply subtype_inc...
         ++ apply Permutation_rev.
      -- rewrite (eta l2) in HAX2 ; inversion HAX2.
         apply inj_pairT2_nat in H1 ; subst.
         rewrite (eta l1) in H2 ; inversion H2.
         apply inj_pairT2_nat in H1 ; subst.
         eapply subtype_inc...
    * assert (exists k', k = S k') as [k' Hk].
      { destruct k ; [exfalso | exists k]... }
      revert l1 l2 HAX1 HAX2 Hi ; rewrite Hk ; clear k Hk ; intros.
      assert (i < S k') as Hi'...
      rewrite <- 3 (nth_order_tl _ Hi' Hi).
      rewrite (eta l1) ; rewrite (eta l2) ; simpl.
      eapply IHi.
      -- rewrite (eta l1) in HAX1 ; inversion HAX1.
         apply inj_pairT2_nat in H1 ; subst.
         rewrite (eta l2) in H3.
         eapply Forall_impl ; [ | apply H3 ].
         clear ; simpl ; intros.
         inversion H.
         apply inj_pairT2_nat in H2 ; subst...
      -- rewrite (eta l2) in HAX2 ; inversion HAX2.
         apply inj_pairT2_nat in H1 ; subst.
         rewrite (eta l1) in H3.
         eapply Forall_impl ; [ | apply H3 ].
         clear ; simpl ; intros.
         inversion H.
         apply inj_pairT2_nat in H2 ; subst...
  + intros ; eapply (inc_Forall (fun z y => subtype [y;a] a z))...
    intros ; eapply subtype_inc...
  + intros ; eapply (inc_Forall (fun z y => subtype [y;a] a z))...
    intros ; eapply subtype_inc...
Qed.


(** * Completeness with respect to BCD *)
Definition interlist := List.fold_right inter Ω.

Lemma interlist_app : forall G D, interlist (G ++ D) ≤ interlist G ∩ interlist D.
Proof with try constructor.
induction G ; intros D.
- eapply trans_bcd.
  + apply inter_right_bcd.
  + apply inter_bcd...
- eapply trans_bcd.
  + rewrite <- app_comm_cons.
    eapply trans_bcd.
    * apply inter_bcd.
      -- apply id_bcd.
      -- apply IHG.
    * eapply trans_bcd.
      -- apply inter_right_bcd.
      -- apply inter_bcd.
         ++ apply inter_bcd.
            ** apply id_bcd.
            ** apply inter_left1_bcd.
         ++ apply inter_left2_bcd.
  + apply inter_bcd...
Qed.

Lemma interlist_app_inv : forall G D, interlist G ∩ interlist D ≤ interlist (G ++ D).
Proof.
induction G ; intros D.
- apply inter_left2_bcd.
- eapply trans_bcd.
  + eapply trans_bcd.
    * apply inter_right_bcd.
    * apply inter_bcd.
      -- eapply trans_bcd.
         ++ apply inter_left1_bcd.
         ++ apply inter_left1_bcd.
      -- apply inter_bcd.
         ++ apply inter_left2_bcd.
         ++ apply id_bcd.
  + rewrite <- app_comm_cons.
    apply inter_bcd.
    * apply id_bcd.
    * apply IHG.
Qed.

Theorem completeness : forall G s n, subtype G s n -> interlist G ≤ s.
Proof with try eassumption ; try reflexivity ; try lia.
intros G s n ; revert G s.
induction n using (well_founded_induction lt_wf) ;
  intros G s pi ; inversion pi ; subst.
- apply H in H0...
  eapply trans_bcd.
  + apply interlist_app.
  + eapply trans_bcd.
    * apply inter_bcd.
      -- apply id_bcd.
      -- apply inter_left2_bcd.
    * eapply trans_bcd.
      -- apply interlist_app_inv.
      -- assumption.
- eapply trans_bcd.
  + apply inter_right_bcd.
  + apply inter_bcd.
    * apply H in H0...
    * apply H in H1...
- apply H in H0...
  eapply trans_bcd.
  + apply interlist_app.
  + eapply trans_bcd.
    * apply inter_bcd.
      -- apply id_bcd.
      -- eapply trans_bcd.
         ++ apply inter_right_bcd.
         ++ apply inter_bcd.
            ** eapply trans_bcd.
               --- apply inter_left1_bcd.
               --- apply inter_left1_bcd.
            ** simpl ; apply inter_bcd.
               --- apply inter_left2_bcd.
               --- apply id_bcd.
    * eapply trans_bcd.
      -- change (s0 ∩ (t ∩ (interlist D))) with (interlist (s0 :: t :: D)).
         apply interlist_app_inv.
      -- assumption.
- destruct C ; [ destruct o | ].
  + assert (forall i (Hi : i < β (Some (Some b0))),
              interlist (List.map (fun c => nth_order (snd c) Hi) L) ≤ nth_order b Hi) as Hb.
    { intros i Hi.
      specialize H2 with i Hi ; simpl in H2.
      apply H in H2... }
    remember (length L) as k.
    clear H1 H2 pi.
    revert L Heqk H0 Hb H ; clear ;
      induction k using (well_founded_induction lt_wf) ;
      intros L Heqk He Hb IH.
    destruct k ; [ | destruct k] ; destruct L ; inversion Heqk.
    * exfalso ; clear - He ; simpl in He...
    * destruct L ; inversion Heqk.
      destruct p.
      simpl in Hb ; simpl ; simpl in t ; simpl in a.
      apply (case0_nil t).
      apply (case0_nil a).
      eapply trans_bcd.
      -- apply inter_left1_bcd.
      -- apply (ctor_mon_bcd_gen _ _ b).
         apply nth_order_Forall2 ; intros i Hi.
         eapply trans_bcd ; [ eapply trans_bcd | ].
         ++ apply inter_right_bcd.
         ++ apply inter_bcd.
            ** apply id_bcd.
            ** apply omega_right_bcd.
         ++ apply Hb...
    * destruct L ; inversion Heqk.
      destruct p ; destruct p0.
      simpl in Hb ; simpl ; simpl in t ; simpl in t1 ; simpl in a.
      apply (case0_nil t).
      apply (case0_nil t1).
      apply (trans_bcd _ (((κ b0 t0) ∩ (κ b0 t2))
                          ∩ interlist (List.map (fun c => Some (Some b0) @ fst c, snd c) L))) ;
        [ eapply trans_bcd | ].
      -- apply inter_right_bcd.
      -- apply inter_bcd.
         ++ apply inter_bcd.
            ** apply id_bcd.
            ** eapply trans_bcd.
               --- apply inter_left1_bcd.
               --- apply id_bcd.
         ++ eapply trans_bcd.
            ** apply inter_left2_bcd.
            ** apply inter_left2_bcd.
      -- apply (trans_bcd _ ((κ b0 (map2 inter t0 t2))
                          ∩ interlist (List.map (fun c => Some (Some b0) @ fst c, snd c) L))).
         ++ apply inter_bcd.
            ** apply ctor_inter_bcd.
            ** apply id_bcd.
         ++ change ((Some (Some b0) @ nil _, map2 inter t0 t2) ∩
                       interlist (List.map (fun c => Some (Some b0) @ fst c, snd c) L))
              with (interlist (List.map (fun c => Some (Some b0) @ fst c, snd c)
                       ((nil _,map2 inter t0 t2) :: L))).
            apply (H (S k)) ; simpl...
            intros i Hi.
            apply (trans_bcd _ ((nth_order t0 Hi) ∩ ((nth_order t2 Hi) ∩
                             interlist (List.map (fun c => nth_order (snd c) Hi) L)))) ;
              [ | apply Hb ]...
            eapply trans_bcd.
            ** apply inter_right_bcd.
            ** apply inter_bcd.
               --- eapply trans_bcd...
                   +++ apply inter_left1_bcd.
                   +++ unfold nth_order.
                       rewrite (nth_map2 _ _ _ _ (Fin.of_nat_lt Hi) (Fin.of_nat_lt Hi))...
                       apply inter_left1_bcd.
               --- apply inter_bcd.
                   +++ unfold nth_order.
                       rewrite (nth_map2 _ _ _ _ (Fin.of_nat_lt Hi) (Fin.of_nat_lt Hi))...
                       apply inter_left2_bcd.
                   +++ apply id_bcd.
  + simpl.
    assert (0 < 1) as HS...
    assert (List.Forall (fun c => nth_order a HS ≤ (nth_order (fst c) HS)) L) as Ha.
    { revert H1 ; clear - H ; induction L ; intro Ha ; constructor.
      - simpl in Ha.
        specialize Ha with 0 HS ; inversion Ha ; subst.
        apply H in H2...
        simpl in H2 ; clear - H2.
        eapply trans_bcd ; [ eapply trans_bcd | ].
        + apply inter_right_bcd.
        + apply inter_bcd.
          * apply id_bcd.
          * apply omega_right_bcd.
        + apply H2.
      - apply IHL...
        intros i0 Hi0 ; simpl in Hi0.
        assert (i0 = 0) by lia ; subst.
        specialize Ha with 0 Hi0 ; inversion Ha... }
    assert (interlist (List.map (fun c => nth_order (snd c) HS) L) ≤ nth_order b HS) as Hb.
    { simpl in H2.
      specialize H2 with 0 HS.
      apply H in H2... }
    clear pi H1 H2 H0.
    remember (length L) as k.
    revert L Heqk HS Ha Hb H ; clear ;
      induction k using (well_founded_induction lt_wf) ;
      intros L Heqk HS Ha Hb IH.
    destruct k ; [ | destruct k] ; destruct L ; inversion Heqk.
    * apply (trans_bcd _ (nth_order a HS → Ω)).
      -- apply (trans_bcd _ (Ω → Ω)).
         ++ apply arrow_omega_bcd.
         ++ apply arrow_bcd.
            ** apply omega_right_bcd.
            ** apply omega_right_bcd.
      -- simpl in a ; simpl in b.
         rewrite (eta a) ; rewrite (eta b) ; simpl.
         rewrite <- hd_nth_order ; simpl.
         apply (case0_nil (tl a)).
         apply (case0_nil (tl b)).
         apply arrow_bcd ; [ apply id_bcd | ].
         simpl in Hb.
         rewrite <- hd_nth_order in Hb...
    * destruct L ; inversion Heqk.
      eapply trans_bcd.
      -- apply inter_left1_bcd.
      -- destruct p ; simpl in t ; simpl in t0 ; simpl in a ; simpl in b.
         rewrite (eta t) ; rewrite (eta t0) ; rewrite (eta a) ; rewrite (eta b) ; simpl.
         apply (case0_nil (tl t)).
         apply (case0_nil (tl t0)).
         apply (case0_nil (tl a)).
         apply (case0_nil (tl b)).
         apply arrow_bcd.
         ++ inversion Ha ; subst.
            rewrite <- 2 (hd_nth_order HS) in H3...
         ++ simpl in Hb.
            eapply trans_bcd ; [ eapply trans_bcd | ].
            ** apply inter_right_bcd.
            ** apply inter_bcd.
               --- apply id_bcd.
               --- apply omega_right_bcd.
            ** rewrite <- 2 (hd_nth_order HS) in Hb.
               apply Hb.
    * destruct L ; inversion Heqk.
      destruct p ; destruct p0 ; simpl.
      apply (trans_bcd _ (((nth_order t HS → nth_order t0 HS) ∩ (nth_order t1 HS → nth_order t2 HS))
                          ∩ interlist (List.map (fun c => Some None @ fst c, snd c) L))).
      -- eapply trans_bcd.
         ++ apply inter_right_bcd.
         ++ apply inter_bcd.
            ** apply inter_bcd.
               --- simpl in t ; simpl in t0 ; rewrite (eta t) ; rewrite (eta t0).
                   rewrite <- 2 (hd_nth_order HS) ; simpl.
                   apply (case0_nil (tl t)).
                   apply (case0_nil (tl t0)).
                   apply id_bcd.
               --- simpl in t1 ; simpl in t2 ; rewrite (eta t1) ; rewrite (eta t2).
                   rewrite <- 2 (hd_nth_order HS) ; simpl.
                   apply (case0_nil (tl t1)).
                   apply (case0_nil (tl t2)).
                   apply inter_left1_bcd.
            ** eapply trans_bcd.
               --- apply inter_left2_bcd.
               --- apply inter_left2_bcd.
      -- apply (trans_bcd _ ((nth_order t HS ∩ nth_order t1 HS → nth_order t0 HS ∩ nth_order t2 HS)
                          ∩ interlist (List.map (fun c => Some None @ fst c, snd c) L))).
         ++ apply inter_bcd.
            ** apply (trans_bcd _ ((nth_order t HS ∩ nth_order t1 HS → nth_order t0 HS)
                                 ∩ (nth_order t HS ∩ nth_order t1 HS → nth_order t2 HS))).
               --- apply inter_bcd.
                   +++ apply arrow_bcd.
                       *** apply inter_left1_bcd.
                       *** apply id_bcd.
                   +++ apply arrow_bcd.
                       *** apply inter_left2_bcd.
                       *** apply id_bcd.
               --- eapply trans_bcd.
                   +++ apply arrow_inter_bcd.
                   +++ apply arrow_bcd ; apply id_bcd.
            ** apply id_bcd.
         ++ change ((nth_order t HS ∩ nth_order t1 HS → nth_order t0 HS ∩ nth_order t2 HS) ∩
                      interlist (List.map (fun c => Some None @ fst c, snd c) L))
              with (interlist (List.map (fun c => Some None @ fst c, snd c)
                      ((cons _ (nth_order t HS ∩ nth_order t1 HS) _ (nil _),
                        cons _ (nth_order t0 HS ∩ nth_order t2 HS) _ (nil _))::L))).
            refine (H (S k) _ _ _ _ _ _ _)...
            ** inversion Ha ; subst.
               inversion H5 ; subst.
               constructor...
               eapply trans_bcd.
               --- apply inter_right_bcd.
               --- apply inter_bcd...
            ** simpl in Hb ; simpl.
               apply (trans_bcd _
                      (nth_order t0 HS ∩ (nth_order t2 HS
                       ∩ interlist (List.map (fun c => nth_order (snd c) HS) L))))...
               --- eapply trans_bcd.
                   +++ apply inter_right_bcd.
                   +++ apply inter_bcd.
                       *** eapply trans_bcd.
                           ---- apply inter_left1_bcd.
                           ---- apply inter_left1_bcd.
                       *** apply inter_bcd.
                           ---- apply inter_left2_bcd.
                           ---- apply id_bcd.
  + clear - a b.
    simpl in L ; simpl in a ; simpl in b ; simpl.
    apply (case0_nil a).
    apply (case0_nil b).
    apply omega_right_bcd.
Qed.

End BCDstd.


